@@ -1,7 +1,7 @@
 """Poisson problem with finite elements
 """
 import numpy as np
-import scipy.sparse as sparse
+from scipy import sparse
 
 
 def check_mesh(V, E):
@@ -353,21 +353,23 @@ def l2norm(u, mesh):
     if mesh.degree == 1:
         I = np.arange(3)
 
-        def basis(x, y):
+        def basis1(x, y):
             return np.array([1-x-y,
                              x,
                              y])
+        basis = basis1
 
     if mesh.degree == 2:
         I = np.arange(6)
 
-        def basis(x, y):
+        def basis2(x, y):
             return np.array([(1-x-y)*(1-2*x-2*y),
                              x*(2*x-1),
                              y*(2*y-1),
                              4*x*(1-x-y),
                              4*x*y,
                              4*y*(1-x-y)])
+        basis = basis2
 
     for e in E:
         x = V[e, 0]
@@ -384,10 +386,13 @@ def l2norm(u, mesh):
     return np.sqrt(val)
 
 
-class mesh:
+class Mesh:
     """Simple mesh object that holds vertices and mesh functions
     """
 
+    # pylint: disable=too-many-instance-attributes
+    # This is reasonble for this class
+
     def __init__(self, V, E, degree=1):
 
         # check to see if E is numbered 0 ... nv
@@ -428,7 +433,8 @@ class mesh:
         """generate a quadratic mesh
         """
         if self.V2 is None:
-            self.V2, self.E2, self.Edges = generate_quadratic(self.V, self.E, return_edges=True)
+            self.V2, self.E2, self.Edges = generate_quadratic(self.V, self.E,
+                                                              return_edges=True)
             self.X2 = self.V2[:, 0]
             self.Y2 = self.V2[:, 1]
             self.newID = self.nv + np.arange(self.Edges.shape[0])
@@ -440,7 +446,7 @@ class mesh:
         self.E2 = None
         self.Edges = None
         self.newID = None
-        for l in range(levels):
+        for _ in range(levels):
             self.V, self.E = refine2dtri(self.V, self.E)
         self.nv = self.V.shape[0]
         self.ne = self.E.shape[0]
@@ -482,22 +488,24 @@ class mesh:
         W = np.array(G.sum(axis=1)).flatten()
 
         Vnew = self.V.copy()
-        edgelength = (Vnew[edge0, 0] - Vnew[edge1, 0])**2 + (Vnew[edge0, 1] - Vnew[edge1, 1])**2
+        edgelength = (Vnew[edge0, 0] - Vnew[edge1, 0])**2 +\
+                     (Vnew[edge0, 1] - Vnew[edge1, 1])**2
 
         maxit = 100
-        for it in range(maxit):
+        for _it in range(maxit):
             Vnew = G @ Vnew
-            Vnew /= W[:, None] # scale the columns by 1/W
-            Vnew[bid,:] = self.V[bid, :]
-            newedgelength = np.sqrt((Vnew[edge0, 0] - Vnew[edge1, 0])**2 +
-                                    (Vnew[edge0, 1] - Vnew[edge1, 1])**2)
+            Vnew /= W[:, None]  # scale the columns by 1/W
+            Vnew[bid, :] = self.V[bid, :]
+            newedgelength = np.sqrt((Vnew[edge0, 0] - Vnew[edge1, 0])**2
+                                    + (Vnew[edge0, 1] - Vnew[edge1, 1])**2)
             move = np.max(np.abs(newedgelength - edgelength) / newedgelength)
             edgelength = newedgelength
             if move < tol:
                 break
 
         self.V = Vnew
-        return it
+        return _it
+
 
 def gradgradform(mesh, kappa=None, f=None, degree=1):
     """Finite element discretization of a Poisson problem.
@@ -585,11 +593,11 @@ def gradgradform(mesh, kappa=None, f=None, degree=1):
         raise ValueError('degree = 1 or 2 supported')
 
     if f is None:
-        def f(x, y):
+        def f(_x, _y):
             return 0.0
 
     if kappa is None:
-        def kappa(x, y):
+        def kappa(_x, _y):
             return 1.0
 
     if not callable(f) or not callable(kappa):
@@ -599,13 +607,11 @@ def gradgradform(mesh, kappa=None, f=None, degree=1):
 
     if degree == 1:
         E = mesh.E
-        V = mesh.V
         X = mesh.X
         Y = mesh.Y
 
     if degree == 2:
         E = mesh.E2
-        V = mesh.V2
         X = mesh.X2
         Y = mesh.Y2
 
@@ -712,7 +718,6 @@ def divform(mesh):
     X, Y = mesh.X, mesh.Y
     ne = mesh.ne
     E = mesh.E2
-    V = mesh.V2
 
     m1 = 6
     m2 = 3
@@ -741,7 +746,7 @@ def divform(mesh):
                        [-0.10810301816806966, -0.10810301816807061],
                        [-0.78379396366386020, -0.10810301816806944],
                        [-0.81684757298045740, -0.81684757298045920],
-                       [ 0.63369514596091700, -0.81684757298045810],
+                       [ 0.63369514596091700, -0.81684757298045810],   # noqa: E201
                        [-0.81684757298045870,  0.63369514596091750]])
         xx, yy = (xy[:, 0]+1)/2, (xy[:, 1]+1)/2
         ww *= 0.5
@@ -752,12 +757,12 @@ def divform(mesh):
         for w, x, y in zip(ww, xx, yy):
             basis1 = np.array([1-x-y, x, y])
 
-            basis2 = np.array([(1-x-y)*(1-2*x-2*y),
-                              x*(2*x-1),
-                              y*(2*y-1),
-                              4*x*(1-x-y),
-                              4*x*y,
-                              4*y*(1-x-y)])
+            # basis2 = np.array([(1-x-y)*(1-2*x-2*y),
+            #                   x*(2*x-1),
+            #                   y*(2*y-1),
+            #                   4*x*(1-x-y),
+            #                   4*x*y,
+            #                   4*y*(1-x-y)])
 
             dbasis = np.array([
                 [4*x + 4*y - 3, 4*x-1,     0, -8*x - 4*y + 4, 4*y,           -4*y],
@@ -766,8 +771,8 @@ def divform(mesh):
 
             dphi = invJ.dot(dbasis)
 
-            DXelem += (detJ / 2) * w * (np.outer(basis1, dphi[0,:]))
-            DYelem += (detJ / 2) * w * (np.outer(basis1, dphi[1,:]))
+            DXelem += (detJ / 2) * w * (np.outer(basis1, dphi[0, :]))
+            DYelem += (detJ / 2) * w * (np.outer(basis1, dphi[1, :]))
             dphi.T.dot(dphi)
 
         # Step 7
@@ -846,9 +851,7 @@ def applybc(A, b, mesh, bc):
         idx = c['var'] + c['id']
         Dflag[idx] = True
     # write identity (2 of 2)
-    for k in range(0, len(A.data)):
-        i = A.row[k]
-        j = A.col[k]
+    for k, (i, j) in enumerate(zip(A.row, A.col)):
         if Dflag[i] or Dflag[j]:
             if i == j:
                 A.data[k] = 1.0
@@ -898,4 +901,5 @@ def model(num=0):
     Notes
     -----
     """
-    pass
+    print(num)
+    raise NotImplementedError('model is unimplemented')
